"""."""

import numpy as _np


def get_matrix_ss_section(section_type):
    SS_LENGTHS = {'SA': 7.0358, 'SB': 6.1758, 'SP': 6.1758}
    length = SS_LENGTHS[section_type]
    return _np.array([
        [1, -length / 2, 0, 0],
        [1, length / 2, 0, 0],
        [0, 0, 1, -length / 2],
        [0, 0, 1, length / 2],
    ])


def si_calculate_bump(
    orbx,
    orby,
    subsec,
    agx=0,
    agy=0,
    psx=0,
    psy=0,
    n_bpms_out=3,
    minsingval=0.2,
):
    """Calculate bumps on SA, SB, SP, BC, C2 and C1 subsections.

    Inputs:
        agx, agy, psx, psy - are the desired angles and positions at the
            center of the straight section/center of BC/source point at B2/
            source point at B1. The must have the same unit as orbx and orby.
        subsec - is a 4 character string with the name of the subsection where
            the bump is to be made. Examples: '01SA', '10C1', '13BC', '02SB'
        orbx, orby - the 160-long vector related to the storage ring orbit.
            Ideally this is the reference orbit around which the bump will be
            made.

    Outputs:
        orbx, orby - The orbits with the bump applied to them
    """
    # get section and subsection
    section_nr = int(subsec[:2])
    if not 1 <= section_nr <= 20:
        raise ValueError('Section must be between 01..20.')

    section_type = subsec[2:]
    if 'S' in section_type:
        mat_s2r = get_matrix_ss_section(section_type)
        idcs = [(section_nr - 1) * 8 - 1, (section_nr - 1) * 8]
    else:
        from apsuite.orbcorr.si_bumps import SiCalcBumps

        bump = SiCalcBumps()
        _, _, mat_s2r = bump.calc_matrices(
            section_type=section_type,
            section_nr=section_nr,
            n_bpms_out=n_bpms_out,
            use_ss_tfm=False,
            minsingval=minsingval,
        )
        idcs = bump.get_bpm_indices(section_type, section_nr - 1)

    vec = _np.array([psx, agx, psy, agy])
    pos_bpm = _np.dot(mat_s2r, vec)

    bpm1 = idcs[0]
    bpm2 = idcs[1]

    orbx, orby = orbx.copy(), orby.copy()
    orbx[bpm1] += pos_bpm[0]
    orbx[bpm2] += pos_bpm[1]
    orby[bpm1] += pos_bpm[2]
    orby[bpm2] += pos_bpm[3]
    return orbx, orby



"""AS RF configuration."""
from copy import deepcopy as _dcopy
import numpy as _np

az = _np.zeros(5)

# NOTE: absolute imports are necessary here due to how
# CONFIG_TYPES in __init__.py is built.
from siriuspy.clientconfigdb.types.global_config import _pvs_as_rf, \
    _pvs_li_llrf


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.

_pvs_bo_llrf = [
    # Interlock disable
    ['RA-RaBO01:RF-LLRF:FIMRevSSA1-Sel', 0.0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRevSSA2-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRevSSA3-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRevSSA4-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRevCav-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMManual-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMPLC-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMLLRF1-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMLLRF2-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMLLRF3-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMPLG1Up-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMPLG1Down-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMPLG2Up-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMPLG2Down-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMCav-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMFwdCav-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMFwdSSA1-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn7-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn8-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn9-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn10-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn11-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn12-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn13-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn14-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMRFIn15-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMOrbitIntlk-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMQuenchCond1-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMEQuench-Sel', 0, 0.0],
    # End switches logic
    ['RA-RaBO01:RF-LLRF:EndSwLogicInv-Sel', 0, 0.0],
    # Beam trip logic
    ['RA-RaBO01:RF-LLRF:OrbitIntlkLogicInv-Sel', 0, 0.0],
    # Vacuum sensor logic
    ['RA-RaBO01:RF-LLRF:VacLogicInv-Sel', 0, 0.0],
    # Pwr interlock threshold
    ['RA-RaBO01:RF-LLRF:LimRevSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRevSSA2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRevSSA3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRevSSA4-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRevCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimFwdCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimFwdSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn7-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn8-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn9-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn10-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn11-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn12-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn13-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn14-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:LimRFIn15-SP', 0, 0.0],
    # Interlock delay
    ['RA-RaBO01:RF-LLRF:IntlkDly-SP', 0, 0.0],
    # Settings PVs values lims
    ### Are we still saving this?
    ['RA-RaBO01:RF-LLRF:ALRef-SP.DRVH', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:ALRef-SP.DRVL', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpTop-SP.DRVH', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpTop-SP.DRVL', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpBot-SP.DRVH', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpBot-SP.DRVL', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:OLGain-SP.DRVH', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:OLGain-SP.DRVL', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:SLKP-SP.DRVH', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:SLKP-SP.DRVL', 0, 0.0],
    # Pressure threshold
    ['BO-05D:VA-CCG-RFC:FastRelay-SP', 0, 0.0],
    # Pressure Lock power increase
    ['RA-RaBO01:RF-LLRF:CondAuto-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:EPSEnbl-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FIMEnbl-Sel', 0, 0.0],
    # ADC Phase and Gain
    ['RA-RaBO01:RF-LLRF:PhShCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShFwdCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShFwdSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShFwdSSA2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShFwdSSA3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShFwdSSA4-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainFwdCav-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainFwdSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainFwdSSA2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainFwdSSA3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainFwdSSA4-SP', 0, 0.0],
    # DAC Phse and Gain
    ['RA-RaBO01:RF-LLRF:PhShSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShSSA2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShSSA3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PhShSSA4-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainSSA1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainSSA2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainSSA3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:GainSSA4-SP', 0, 0.0],
    # Loops parameters
    ['RA-RaBO01:RF-LLRF:SLKP-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:SLKI-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:SLPILim-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:SLInp-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FLKP-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FLKI-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FLPILim-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FLInp-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:ALKP-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:ALKI-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:ALInp-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PLKP-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PLKI-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:PLInp-Sel', 0, 0.0],
    # Loop mode
    ['RA-RaBO01:RF-LLRF:LoopMode-Sel', 0, 0.0],
    # Min forward power
    ['RA-RaBO01:RF-LLRF:LoopFwdMin-SP', 0, 0.0],
    # Min amplitude ref
    ['RA-RaBO01:RF-LLRF:AmpRefMin-SP', 0, 0.0],
    # Min phase ref
    ['RA-RaBO01:RF-LLRF:PhsRefMin-SP', 0, 0.0],
    # Open loop gain
    ['RA-RaBO01:RF-LLRF:OLGain-SP', 0, 0.0],
    # Tuning loop config
    ['RA-RaBO01:RF-LLRF:Tune-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneDir-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneFwdMin-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneMarginHI-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneMarginLO-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Detune-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneDly-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneFreq-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneFilt-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:TuneTrig-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpTuneTop-Sel', 0, 0.0],
    # Field Flatness loop config
    ['RA-RaBO01:RF-LLRF:FFEnbl-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FFDir-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FFDeadBand-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FFGainCell2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FFGainCell4-SP', 0, 0.0],
    # Pulsed mode config
    ['RA-RaBO01:RF-LLRF:CondFreq-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:CondDuty2-SP', 0, 0.0],
    # Rmp mode config
    ['RA-RaBO01:RF-LLRF:RmpTs1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpTs2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpTs3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpTs4-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpTop-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpPhsTop-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpAmpBot-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpPhsBot-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpIncTime-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RmpDownDsbl-Sel', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FDLFrame-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FDLRearm-Sel', 0, 0.0],
    # Enable plungers step motor drivers
    ['RA-RaBO01:RF-CavPlDrivers:DrEnbl-Sel', 0, 0.0],
    # Enable ADC phase and gain
    ['RA-RaBO01:RF-LLRF:PhShADC-Sel', 0, 0.0],
    # Enable DAC phase and gain
    ['RA-RaBO01:RF-LLRF:PhShDAC-Sel', 0, 0.0]
    ]

_pvs_bo_rfssa = [
    # SSA tower offsets
    ['RA-ToBO:OffsetConfig:UpperIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:UpperReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:LowerIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:LowerReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:InputIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:InputReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:OutputIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:OutputReflectedPower', 0, 0.0],  # [dB]
    # SSA tower pwr alarm limits
    ['RA-ToBO:AlarmConfig:GeneralPowerLimHiHi', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimHigh', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimLow', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimLoLo', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimHigh', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimLow', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],  # [dBm]
    # SSA tower current alarm limits
    ['RA-ToBO:AlarmConfig:CurrentLimHiHi', 0, 0.0],  # [A]
    ['RA-ToBO:AlarmConfig:CurrentLimHigh', 0, 0.0],  # [A]
    ['RA-ToBO:AlarmConfig:CurrentLimLow', 0, 0.0],  # [A]
    ['RA-ToBO:AlarmConfig:CurrentLimLoLo', 0, 0.0],  # [A]
    ]

_pvs_bo_rfcal = [
    ['RA-RaBO01:RF-LLRF:CavOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:CavSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:CavSysCalInv-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdCavOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdCavSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdCavSysCalInv-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:RevCavOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RevCavSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:MOOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:MOSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdSSA1Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdSSA1SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdSSA1SysCalInv-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:RevSSA1Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RevSSA1SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell2Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell2SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell4Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell4SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell1Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell1SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell5Offset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Cell5SysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:InPre1AmpOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:InPre1AmpSysCal-SP', az, 0.0],


    ['RA-RaBO01:RF-LLRF:RevPreAmpOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RevPreAmpSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdCircOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:FwdCircSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:RevCircOffset-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:RevCircSysCal-SP', az, 0.0],
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff-SP', az, 0.0],
    ['BO-05D:RF-P5Cav:Rsh-Cte', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff-SP', az, 0.0],
    # CalSys Offsets
    ['RA-RaBO01:RF-RFCalSys:OFSdB1-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB2-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB3-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB4-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB5-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB6-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB7-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB8-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB9-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB10-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB11-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB12-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB13-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB14-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB15-Mon', 0, 0.0],
    ['RA-RaBO01:RF-RFCalSys:OFSdB16-Mon', 0, 0.0]
]

_pvs_bo_pow_sensor = [
    #Keysight U2021xa Power Sensor config
    ['RA-RF:PowerSensor1:GainOffsetStat-Sel', 0, 0.0],
    ['RA-RF:PowerSensor1:GainOffset-SP', 0, 0.0],
    ['RA-RF:PowerSensor1:Egu-SP', 0, 0.0],
    ['RA-RF:PowerSensor1:TracTime-SP', 0, 0.0],
    ['RA-RF:PowerSensor1:Freq-SP', 0, 0.0],
    ]

##SIA PVs

_pvs_sia_llrf = [
    # Interlock disable
    ['RA-RaSIA01:RF-LLRF:FIMRevSSA1-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRevSSA2-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRevSSA3-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRevSSA4-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRevCav-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMManual-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMPLC-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMLLRF1-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMLLRF2-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMLLRF3-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMTunerHigh-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMTunerLow-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMPLG2Up-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMPLG2Down-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMCav-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMFwdCav-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMFwdSSA1-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn7-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn8-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn9-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn10-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn11-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn12-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn13-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn14-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMRFIn15-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMOrbitIntlk-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMQuenchCond1-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMEQuench-Sel', 0, 0.0],
    # End switches logic
    ['RA-RaSIA01:RF-LLRF:EndSwLogicInv-Sel', 0, 0.0],
    # Beam trip logic
    ['RA-RaSIA01:RF-LLRF:OrbitIntlkLogicInv-Sel', 0, 0.0],
    # Vacuum sensor logic
    ['RA-RaSIA01:RF-LLRF:VacLogicInv-Sel', 0, 0.0],
    # Pwr interlock threshold
    ['RA-RaSIA01:RF-LLRF:LimRevSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRevSSA2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRevSSA3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRevSSA4-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRevCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimFwdCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn7-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn8-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn9-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn10-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn11-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn12-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn13-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn14-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:LimRFIn15-SP', 0, 0.0],
    # Interlock delay
    ['RA-RaSIA01:RF-LLRF:IntlkDly-SP', 0, 0.0],
    # Set PVs value lims
    ['RA-RaSIA01:RF-LLRF:ALRef-SP.DRVH', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:ALRef-SP.DRVL', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:OLGain-SP.DRVH', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:OLGain-SP.DRVL', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:SLKP-SP.DRVH', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:SLKP-SP.DRVL', 0, 0.0],
    # Pressure Lock power increase
    ['RA-RaSIA01:RF-LLRF:CondAuto-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:EPSEnbl-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FIMEnbl-Sel', 0, 0.0],
    # ADC Phase and Gain
    ['RA-RaSIA01:RF-LLRF:PhShCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShFwdCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShFwdSSA2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShFwdSSA3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShFwdSSA4-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainFwdCav-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainFwdSSA2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainFwdSSA3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainFwdSSA4-SP', 0, 0.0],
    # DAC Phse and Gain
    ['RA-RaSIA01:RF-LLRF:PhShSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShSSA2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShSSA3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PhShSSA4-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainSSA1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainSSA2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainSSA3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:GainSSA4-SP', 0, 0.0],
    # Loops parameters
    ['RA-RaSIA01:RF-LLRF:SLKP-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:SLKI-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:SLPILim-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:SLInp-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FLKP-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FLKI-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FLPILim-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FLInp-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:ALKP-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:ALKI-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:ALInp-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PLKP-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PLKI-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:PLInp-Sel', 0, 0.0],
    # Loop mode
    ['RA-RaSIA01:RF-LLRF:LoopMode-Sel', 0, 0.0],
    # Min forward power
    ['RA-RaSIA01:RF-LLRF:LoopFwdMin-SP', 0, 0.0],
    # Min amplitude reference
    ['RA-RaSIA01:RF-LLRF:AmpRefMin-SP', 0, 0.0],
    # Min phase reference
    ['RA-RaSIA01:RF-LLRF:PhsRefMin-SP', 0, 0.0],
    # Open loop gain
    ['RA-RaSIA01:RF-LLRF:OLGain-SP', 0, 0.0],
    # Phase ref
    ['RA-RaSIA01:RF-LLRF:PLRef-SP', 0, 0.0],
    # Tuning loop config
    ['RA-RaSIA01:RF-LLRF:Tune-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneDir-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneFwdMin-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneMarginHI-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneMarginLO-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:Detune-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneDly-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneFreq-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneFilt-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:TuneTrig-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpTuneTop-Sel', 0, 0.0],
    # Rmp mode config
    ['RA-RaSIA01:RF-LLRF:RmpTs1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpTs2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpTs3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpTs4-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpAmpTop-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpPhsTop-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpAmpBot-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpPhsBot-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpIncTime-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RmpDownDsbl-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FDLFrame-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FDLRearm-Sel', 0, 0.0],
    # Enable ADC phase and gain
    ['RA-RaSIA01:RF-LLRF:PhShADC-Sel', 0, 0.0],
    # Enable DAC phase and gain
    ['RA-RaSIA01:RF-LLRF:PhShDAC-Sel', 0, 0.0],
    # Dac5 outputs
    ['RA-RaSIA01:RF-LLRF:CavMonGain-Sel', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:CavMonCh-Sel', 0, 0.0],
    # Quench Condition
    ['RA-RaSIA01:RF-LLRF:QuenchCond1RvRatio-SP',0, 0.0],
    ['RA-RaSIA01:RF-LLRF:QuenchCond1Dly-SP',0, 0.0],
    # EQuench
    ['RA-RaSIA01:RF-LLRF:EQuenchFwRatio-SP',0, 0.0],
    ['RA-RaSIA01:RF-LLRF:EQuenchDly-SP',0, 0.0],
    # Dynamic Interlock setpoints
    ['RA-RaSIA01:RF-LLRF:CurrentDelta-SP', 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimRevCavCoeff-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimFwdCavCoeff-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:EQuenchFwRatioCoeff-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:QuenchCond1RvRatioCoeff-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimRevCavOffset-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimFwdCavOffset-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:EQuenchFwRatioOffset-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:QuenchCond1RvRatioOffset-SP", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimRevCavEn-Sel", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:LimFwdCavEn-Sel", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:EQuenchFwRatioEn-Sel", 0, 0.0],
    ["RA-RaSIA01:RF-LLRF:QuenchCond1RvRatioEn-Sel", 0, 0.0],
]

_pvs_sia_rfssa = [
    # NOTE: Alarms and offset of SSA towers 3 & 4 temporaly removed
    # SSA tower 1 offsets
    ['RA-ToSIA01:OffsetConfig:UpperIncidentPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # SSA tower 1 pwr alarm limits
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # SSA tower 1 current alarm limits
    ['RA-ToSIA01:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # SSA tower 2 offsets
    ['RA-ToSIA02:OffsetConfig:UpperIncidentPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # SSA tower 2 pwr alarm limits
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # SSA tower 2 current alarm limits
    ['RA-ToSIA02:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # # SSA tower 3 offsets
    # ['RA-ToSIA03:OffsetConfig:UpperIncidentPower', 0, 0.0],
    # ['RA-ToSIA03:OffsetConfig:UpperReflectedPower', 0, 0.0],
    # ['RA-ToSIA03:OffsetConfig:LowerIncidentPower', 0, 0.0],
    # ['RA-ToSIA03:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # # SSA tower 3 pwr alarm limits
    # ['RA-ToSIA03:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # # SSA tower 3 current alarm limits
    # ['RA-ToSIA03:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:CurrentLimHigh', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:CurrentLimLow', 0, 0.0],
    # ['RA-ToSIA03:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # # SSA tower 4 offsets
    # ['RA-ToSIA04:OffsetConfig:UpperIncidentPower', 0, 0.0],
    # ['RA-ToSIA04:OffsetConfig:UpperReflectedPower', 0, 0.0],
    # ['RA-ToSIA04:OffsetConfig:LowerIncidentPower', 0, 0.0],
    # ['RA-ToSIA04:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # # SSA tower 4 pwr alarm limits
    # ['RA-ToSIA04:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # # SSA tower 4 current alarm limits
    # ['RA-ToSIA04:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:CurrentLimHigh', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:CurrentLimLow', 0, 0.0],
    # ['RA-ToSIA04:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    #SSA1 Pwr Cal Coeff
    ['RA-ToSIA01:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    ['RA-ToSIA01:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    ['RA-ToSIA01:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    ['RA-ToSIA01:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    #SSA2 Pwr Cal Coeff
    ['RA-ToSIA02:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    ['RA-ToSIA02:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    ['RA-ToSIA02:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    ['RA-ToSIA02:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    # # SSA3 Pwr Cal Coeff
    # ['RA-ToSIA03:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    # ['RA-ToSIA03:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    # ['RA-ToSIA03:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    # ['RA-ToSIA03:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    # # SSA4 Pwr Cal Coeff
    # ['RA-ToSIA04:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    # ['RA-ToSIA04:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    # ['RA-ToSIA04:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    # ['RA-ToSIA04:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    ]

## Commented out temporarily in case we need to switch back to the Petra 7 cavity.
_pvs_sia_rfcav = [
#     # CavP7 water flow rate
#     ['SI-02SB:RF-P7Cav:Disc1FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell1FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc2FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell2FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc3FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell3FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc4FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell4FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc5FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell5FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc6FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell6FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc7FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Cell7FlwRt-Mon', 0, 0.0],  # [L/h]
#     ['SI-02SB:RF-P7Cav:Disc8FlwRt-Mon', 0, 0.0],  # [L/h]
    ]

_pvs_sia_rfcal = [
    ['RA-RaSIA01:RF-LLRF:CavOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:CavSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:CavSysCalInv-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdCavOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdCavSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdCavSysCalInv-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdLoadOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdLoadSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FBTntOffset-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FBTntSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdInCircOffset-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdInCircSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevInCircOffset-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevInCircSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA1Offset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA1SysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA1SysCalInv-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA2Offset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA2SysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:FwdSSA2SysCalInv-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:In1PreAmpOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:In1PreAmpSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:In2PreAmpOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:In2PreAmpSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:MOOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:MOSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevCavOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevCavSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevLoadOffset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevLoadSysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevSSA1Offset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevSSA1SysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevSSA2Offset-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:RevSSA2SysCal-SP', az, 0.0],
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff-SP', az, 0.0],
    ['SI-03SP:RF-SRFCav-A:Rsh-Cte', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB1-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB2-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB3-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB4-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB5-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB6-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB7-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB8-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB9-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB10-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB11-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB12-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB13-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB14-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB15-Mon', 0, 0.0],
    ['RA-RaSIA01:RF-RFCalSys:OFSdB16-Mon', 0, 0.0]
]

##SIB PVs

_pvs_sib_llrf = [
    # Interlock disable
    ['RA-RaSIB01:RF-LLRF:FIMRevSSA1-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRevSSA2-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRevSSA3-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRevSSA4-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRevCav-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMManual-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMPLC-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMLLRF1-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMLLRF2-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMLLRF3-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMTunerHigh-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMTunerLow-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMPLG2Up-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMPLG2Down-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMCav-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMFwdCav-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMFwdSSA1-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn7-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn8-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn9-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn10-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn11-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn12-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn13-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn14-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMRFIn15-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMOrbitIntlk-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMQuenchCond1-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMEQuench-Sel', 0, 0.0],
    # End switches logic
    ['RA-RaSIB01:RF-LLRF:EndSwLogicInv-Sel', 0, 0.0],
    # Beam trip logic
    ['RA-RaSIB01:RF-LLRF:OrbitIntlkLogicInv-Sel', 0, 0.0],
    # Vacuum sensor logic
    ['RA-RaSIB01:RF-LLRF:VacLogicInv-Sel', 0, 0.0],
    # Pwr interlock threshold
    ['RA-RaSIB01:RF-LLRF:LimRevSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRevSSA2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRevSSA3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRevSSA4-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRevCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimFwdCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn7-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn8-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn9-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn10-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn11-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn12-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn13-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn14-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:LimRFIn15-SP', 0, 0.0],
    # Interlock delay
    ['RA-RaSIB01:RF-LLRF:IntlkDly-SP', 0, 0.0],
    # Set PVs value lims
    ###Are we still saving this?
    ['RA-RaSIB01:RF-LLRF:ALRef-SP.DRVL', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:ALRef-SP.DRVH', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:OLGain-SP.DRVH', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:OLGain-SP.DRVL', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:SLKP-SP.DRVH', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:SLKP-SP.DRVL', 0, 0.0],
    # Pressure Lock power increase
    ['RA-RaSIB01:RF-LLRF:CondAuto-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:EPSEnbl-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FIMEnbl-Sel', 0, 0.0],
    # ADC Phase and Gain
    ['RA-RaSIB01:RF-LLRF:PhShCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShFwdCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShFwdSSA2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShFwdSSA3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShFwdSSA4-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainFwdCav-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainFwdSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainFwdSSA2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainFwdSSA3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainFwdSSA4-SP', 0, 0.0],
    # DAC Phse and Gain
    ['RA-RaSIB01:RF-LLRF:PhShSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShSSA2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShSSA3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PhShSSA4-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainSSA1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainSSA2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainSSA3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:GainSSA4-SP', 0, 0.0],
    # Loops parameters
    ['RA-RaSIB01:RF-LLRF:SLKP-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:SLKI-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:SLPILim-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:SLInp-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FLKP-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FLKI-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FLPILim-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FLInp-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:ALKP-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:ALKI-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:ALInp-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PLKP-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PLKI-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:PLInp-Sel', 0, 0.0],
    # Loop mode
    ['RA-RaSIB01:RF-LLRF:LoopMode-Sel', 0, 0.0],
    # Min forward power
    ['RA-RaSIB01:RF-LLRF:LoopFwdMin-SP', 0, 0.0],
    # Min amplitude reference
    ['RA-RaSIB01:RF-LLRF:AmpRefMin-SP', 0, 0.0],
    # Min phase reference
    ['RA-RaSIB01:RF-LLRF:PhsRefMin-SP', 0, 0.0],
    # Open loop gain
    ['RA-RaSIB01:RF-LLRF:OLGain-SP', 0, 0.0],
    # Phase ref
    ['RA-RaSIB01:RF-LLRF:PLRef-SP', 0, 0.0],
    # Tuning loop config
    ['RA-RaSIB01:RF-LLRF:Tune-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneDir-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneFwdMin-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneMarginHI-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneMarginLO-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:Detune-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneDly-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneFreq-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneFilt-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:TuneTrig-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpTuneTop-Sel', 0, 0.0],
    # Rmp mode config
    ['RA-RaSIB01:RF-LLRF:RmpTs1-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpTs2-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpTs3-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpTs4-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpAmpTop-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpPhsTop-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpAmpBot-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpPhsBot-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpIncTime-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RmpDownDsbl-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FDLFrame-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FDLRearm-Sel', 0, 0.0],
    # Enable ADC phase and gain
    ['RA-RaSIB01:RF-LLRF:PhShADC-Sel', 0, 0.0],
    # Enable DAC phase and gain
    ['RA-RaSIB01:RF-LLRF:PhShDAC-Sel', 0, 0.0],
    # DAC5 Output
    ['RA-RaSIB01:RF-LLRF:CavMonGain-Sel', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:CavMonCh-Sel', 0, 0.0],
    # Quench Condition
    ['RA-RaSIB01:RF-LLRF:QuenchCond1RvRatio-SP',0, 0.0],
    ['RA-RaSIB01:RF-LLRF:QuenchCond1Dly-SP',0, 0.0],
    # EQuench
    ['RA-RaSIB01:RF-LLRF:EQuenchFwRatio-SP',0, 0.0],
    ['RA-RaSIB01:RF-LLRF:EQuenchDly-SP',0, 0.0],
    # Dynamic Interlock setpoints
    ['RA-RaSIB01:RF-LLRF:CurrentDelta-SP', 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimRevCavCoeff-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimFwdCavCoeff-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:EQuenchFwRatioCoeff-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:QuenchCond1RvRatioCoeff-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimRevCavOffset-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimFwdCavOffset-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:EQuenchFwRatioOffset-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:QuenchCond1RvRatioOffset-SP", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimRevCavEn-Sel", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:LimFwdCavEn-Sel", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:EQuenchFwRatioEn-Sel", 0, 0.0],
    ["RA-RaSIB01:RF-LLRF:QuenchCond1RvRatioEn-Sel", 0, 0.0],
]

_pvs_sib_rfssa = [
    # SSA tower 1 offsets
    # ['RA-ToSIB01:OffsetConfig:UpperIncidentPower', 0, 0.0],
    # ['RA-ToSIB01:OffsetConfig:UpperReflectedPower', 0, 0.0],
    # ['RA-ToSIB01:OffsetConfig:LowerIncidentPower', 0, 0.0],
    # ['RA-ToSIB01:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # # SSA tower 1 pwr alarm limits
    # ['RA-ToSIB01:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # # SSA tower 1 current alarm limits
    # ['RA-ToSIB01:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:CurrentLimHigh', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:CurrentLimLow', 0, 0.0],
    # ['RA-ToSIB01:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # # SSA tower 2 offsets
    # ['RA-ToSIB02:OffsetConfig:UpperIncidentPower', 0, 0.0],
    # ['RA-ToSIB02:OffsetConfig:UpperReflectedPower', 0, 0.0],
    # ['RA-ToSIB02:OffsetConfig:LowerIncidentPower', 0, 0.0],
    # ['RA-ToSIB02:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # # SSA tower 2 pwr alarm limits
    # ['RA-ToSIB02:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # # SSA tower 2 current alarm limits
    # ['RA-ToSIB02:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:CurrentLimHigh', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:CurrentLimLow', 0, 0.0],
    # ['RA-ToSIB02:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # SSA tower 3 offsets
    ['RA-ToSIB03:OffsetConfig:UpperIncidentPower', 0, 0.0],
    ['RA-ToSIB03:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIB03:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIB03:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # SSA tower 3 pwr alarm limits
    ['RA-ToSIB03:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # SSA tower 3 current alarm limits
    ['RA-ToSIB03:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIB03:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # SSA tower 4 offsets
    ['RA-ToSIB04:OffsetConfig:UpperIncidentPower', 0, 0.0],
    ['RA-ToSIB04:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIB04:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIB04:OffsetConfig:LowerReflectedPower', 0, 0.0],
    # SSA tower 4 pwr alarm limits
    ['RA-ToSIB04:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    # SSA tower 4 current alarm limits
    ['RA-ToSIB04:AlarmConfig:CurrentLimHiHi', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIB04:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    # #SSA1 Pwr Cal Coeff
    # ['RA-ToSIB01:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    # ['RA-ToSIB01:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    # ['RA-ToSIB01:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    # ['RA-ToSIB01:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    # #SSA2 Pwr Cal Coeff
    # ['RA-ToSIB02:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    # ['RA-ToSIB02:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    # ['RA-ToSIB02:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    # ['RA-ToSIB02:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    # SSA3 Pwr Cal Coeff
    ['RA-ToSIB03:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    ['RA-ToSIB03:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    ['RA-ToSIB03:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    ['RA-ToSIB03:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    # SSA4 Pwr Cal Coeff
    ['RA-ToSIB04:RF-SSAmpTower:Hw2PwrFwdInCoeff-Cte', az, 0.0],
    ['RA-ToSIB04:RF-SSAmpTower:Hw2PwrRevInCoeff-Cte', az, 0.0],
    ['RA-ToSIB04:RF-SSAmpTower:Hw2PwrFwdOutCoeff-Cte', az, 0.0],
    ['RA-ToSIB04:RF-SSAmpTower:Hw2PwrRevOutCoeff-Cte', az, 0.0],
    ]

_pvs_sib_rfcav = []

_pvs_sib_rfcal = [
    ['RA-RaSIB01:RF-LLRF:CavOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:CavSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:CavSysCalInv-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdCavOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdCavSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdCavSysCalInv-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdLoadOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdLoadSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FBTntOffset-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FBTntSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdInCircOffset-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdInCircSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevInCircOffset-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevInCircSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA1Offset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA1SysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA1SysCalInv-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA2Offset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA2SysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:FwdSSA2SysCalInv-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:In1PreAmpOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:In1PreAmpSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:In2PreAmpOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:In2PreAmpSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:MOOffset-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:MOSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevCavOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevCavSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevLoadOffset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevLoadSysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevSSA1Offset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevSSA1SysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevSSA2Offset-SP', 0, 0.0],
    ['RA-RaSIB01:RF-LLRF:RevSSA2SysCal-SP', az, 0.0],
    ['RA-RaSIB01:RF-LLRF:AmpVCav2HwCoeff-SP', az, 0.0],
    ['SI-03SP:RF-SRFCav-B:Rsh-Cte', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB1-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB2-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB3-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB4-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB5-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB6-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB7-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB8-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB9-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB10-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB11-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB12-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB13-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB14-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB15-Mon', 0, 0.0],
    ['RA-RaSIB01:RF-RFCalSys:OFSdB16-Mon', 0, 0.0]
]

##
_template_dict = {
    'pvs':
    _pvs_as_rf + _pvs_li_llrf +
    _pvs_bo_pow_sensor + _pvs_bo_llrf + _pvs_bo_rfssa + _pvs_bo_rfcal +
    _pvs_sia_llrf + _pvs_sia_rfssa + _pvs_sia_rfcav + _pvs_sia_rfcal +
    _pvs_sib_llrf + _pvs_sib_rfssa + _pvs_sib_rfcav + _pvs_sib_rfcal
    }



"""Beam Loss Monitor devices."""

import numbers as _numbers

import numpy as _np

from ..search import GammaMonitorSearch as _GammaMonitorSearch
from ..util import ClassProperty as _classproperty
from .device import Device as _Device, DeviceSet as _DeviceSet


class GammaMonitor(_Device):
    """Gamma Monitor Device."""

    class DEVICES:
        """Devices names."""
        __NAMES = None

        @_classproperty
        def ALL(cls):  # noqa: N802, N805
            """."""
            if cls.__NAMES is None:
                cls.__NAMES = _GammaMonitorSearch.get_gammanames()
            return cls.__NAMES

    PROPERTIES_DEFAULT = ('Count-Mon',)

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in GammaMonitor.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def counts(self):
        """Return Gamma Monitor count value [pulse/s]."""
        return self['Count-Mon']


class GammaCounter(_Device):
    """Gamma Monitor Counter Device."""

    class DEVICES:
        """Devices names."""
        __NAMES = None

        @_classproperty
        def ALL(cls):  # noqa: N802, N805
            """."""
            if cls.__NAMES is None:
                cls.__NAMES = _GammaMonitorSearch.get_counter_names()
            return cls.__NAMES

    PROPERTIES_DEFAULT = ('TimeBase-SP',)

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in GammaCounter.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def time_base(self):
        """Return Gamma Monitor Counter time base value."""
        return self['TimeBase-SP']

    @time_base.setter
    def time_base(self, value):
        """Set Gamma Monitor Counter time base value."""
        self['TimeBase-SP'] = int(value)


class FamGammaMonitors(_DeviceSet):
    """Gamma Monitor Device Set."""

    def __init__(self, props2init='all'):
        """."""
        self.monitors = [
            GammaMonitor(devname, props2init=props2init)
            for devname in GammaMonitor.DEVICES.ALL
        ]
        self.counters = [
            GammaCounter(devname, props2init=props2init)
            for devname in GammaCounter.DEVICES.ALL
        ]
        super().__init__(self.monitors + self.counters)

    @property
    def counts(self):
        """Return list of Gamma Monitor count values [pulse/s]."""
        return _np.array([dev.counts for dev in self.monitors])

    @property
    def time_bases(self):
        """Return list of Gamma Monitor Counter time base values."""
        return _np.array([counter.time_base for counter in self.counters])

    @time_bases.setter
    def time_bases(self, value):
        """Set list of Gamma Monitor Counter time base values."""
        if isinstance(value, _numbers.Number):
            value = [value] * len(self.counters)
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise ValueError(
                'Value must be a number, list, tuple or numpy array.'
            )
        if len(value) != len(self.counters):
            raise ValueError(
                f'Value must have the same length as the number of counters '
                f'({len(self.counters)}).'
            )
        for counter, time_base in zip(self.counters, value):
            counter.time_base = time_base

    @property
    def monitor_names(self):
        """Return list of Gamma Monitor names."""
        return [dev.devname for dev in self.monitors]

    @property
    def counter_names(self):
        """Return list of Gamma Monitor Counter names."""
        return [dev.devname for dev in self.counters]



"""DVF devices."""

import numpy as _np
from mathphys.functions import get_namedtuple as _get_namedtuple
from mathphys.imgproc import (
    FitGaussianScipy as _FitGaussianScipy,
    Image2D_Fit as _Image2D_Fit
)

from .device import Device as _Device


class DVF(_Device):
    """Beam Visualization Device ("Dispositivo de Visualização de Feixe")."""

    class DEVICES:
        """Devices names."""

        CAX_DVF1 = 'CAX:A:BASLER01'
        CAX_DVF2 = 'CAX:B:BASLER01'
        EMA_DVF2 = 'EMA:A:BASLER02'
        PNR_DVF2 = 'PNR:A:BASLER02'
        BO_DVF = 'BO-50U:DI-VLightCam'
        ALL = (CAX_DVF1, CAX_DVF2, BO_DVF, EMA_DVF2, PNR_DVF2)

    _default_timeout = 10  # [s]

    _dvfparam_fields = (
        'MAX_INTENSITY_NR_BITS',
        'ACQUISITION_TIME_MIN',  # [s]
        'ACQUISITION_TIME_DEFAULT',  # [s]
        'EXPOSURE_TIME_DEFAULT',  # [s]
        'IMAGE_SIZE_Y',  # [pixel]
        'IMAGE_SIZE_X',  # [pixel]
        'IMAGE_PIXEL_SIZE',  # [um]
        'OPTICS_MAGNIFICATION_FACTOR',  # source to image
    )

    _dev2params = {
        DEVICES.CAX_DVF1: _get_namedtuple(
            'DVFParameters',
            _dvfparam_fields,
            (16, 0.5, 0.5, 0.100, 2064, 3088, 2.4, 5.0),
        ),
        DEVICES.CAX_DVF2: _get_namedtuple(
            'DVFParameters',
            _dvfparam_fields,
            (16, 0.5, 0.5, 0.100, 2064, 3088, 2.4, 5.0),
        ),
        DEVICES.BO_DVF: _get_namedtuple(
            'DVFParameters',
            _dvfparam_fields,
            (8, 0.5, 0.5, 0.005, 1024, 1280, 4.8, 5.0),
        ),
    }

    PROPERTIES_DEFAULT = (
        'cam1:MaxSizeX_RBV',
        'cam1:MaxSizeY_RBV',
        'cam1:SizeX',
        'cam1:SizeX_RBV',
        'cam1:SizeY',
        'cam1:SizeY_RBV',
        'cam1:MinX',
        'cam1:MinX_RBV',
        'cam1:MinY',
        'cam1:MinY_RBV',
        'cam1:PortName_RBV',
        'cam1:ArrayCallbacks',
        'cam1:ArrayCallbacks_RBV',
        'cam1:AcquireTime',
        'cam1:AcquireTime_RBV',
        'cam1:AcquirePeriod',
        'cam1:AcquirePeriod_RBV',
        'cam1:Acquire',
        'cam1:Acquire_RBV',
        'cam1:ImageMode',
        'cam1:ImageMode_RBV',
        'cam1:Gain',
        'cam1:Gain_RBV',
        'cam1:GainAuto',
        'cam1:GainAuto_RBV',
        'cam1:PixelFormat',
        'cam1:PixelFormat_RBV',
        'cam1:GC_TemperatureAbs_RBV',
        'cam1:ARFrameFailures',
        'cam1:ARFramesCompleted',
        'image1:NDArrayPort',
        'image1:NDArrayPort_RBV',
        'image1:EnableCallbacks',
        'image1:EnableCallbacks_RBV',
        'image1:ArraySize0_RBV',
        'image1:ArraySize1_RBV',
        'image1:DataType_RBV',
        'image1:ArrayData',
        'ROI1:PortName_RBV',
        'ROI1:NDArrayPort',
        'ROI1:NDArrayPort_RBV',
        'ROI1:EnableCallbacks',
        'ROI1:EnableCallbacks_RBV',
        'ROI1:MinX',
        'ROI1:MinX_RBV',
        'ROI1:MinY',
        'ROI1:MinY_RBV',
        'ROI1:SizeX',
        'ROI1:SizeX_RBV',
        'ROI1:SizeY',
        'ROI1:SizeY_RBV',
        'ROI1:EnableX',
        'ROI1:EnableX_RBV',
        'ROI1:EnableY',
        'ROI1:EnableY_RBV',
        'ROI1:ArrayCallbacks',
        'ROI1:ArrayCallbacks_RBV',
        'Trans1:EnableCallbacks',
        'Trans1:EnableCallbacks_RBV',
    )

    _PROPTY_NAME_MAP = dict()

    def __init__(self, devname, props2init='all', **kwargs):
        """Init."""
        if devname not in DVF.DEVICES.ALL:
            raise NotImplementedError(devname)
        if devname == DVF.DEVICES.BO_DVF:
            self._devname = devname
            props = self.PROPERTIES_DEFAULT
            self.PROPERTIES_DEFAULT = [
                self._get_propty(prop) for prop in props
            ]
        super().__init__(devname, props2init=props2init, **kwargs)

    @property
    def parameters(self):
        """Return DVF parameters."""
        return DVF._dev2params[self.devname]

    @property
    def intensity_saturation_value(self):
        """Image intensity saturation value."""
        pixel_format_to_num_bits = {
            0: 8,  # Mono8
            1: 12,  # Mono12
            2: 12,  # Mono12Packed
        }
        data_type_to_num_bits = {
            0: 8,  # Int8
            1: 8,  # UInt8
            2: 16,  # Int16
            3: 16,  # Uint16
            4: 32,  # Int32
            5: 32,  # UInt32
            6: 64,  # Int64
            7: 64,  # UInt64
            8: 32,  # Float32
            9: 64,  # Float64
        }
        used_bits = pixel_format_to_num_bits[self.pixel_format]
        max_bits = data_type_to_num_bits[self.data_type]
        max_intensity = (1 << used_bits) - 1
        if used_bits <= max_bits:
            # Shift intensity to align with most-significant bits
            max_intensity <<= max_bits - used_bits
        elif used_bits > max_bits:
            # Clip to maximum value allowed by the data type
            max_intensity = (1 << max_bits) - 1
        return max_intensity

    @property
    def exposure_time(self):
        """Camera exposure time [s]."""
        return self['cam1:AcquireTime']

    @exposure_time.setter
    def exposure_time(self, value):
        """Camera exposure time [s]."""
        value = max(value, 0)
        self['cam1:AcquireTime'] = value

    @property
    def acquisition_time(self):
        """Camera acquisition time [s]."""
        return self['cam1:AcquirePeriod']

    @acquisition_time.setter
    def acquisition_time(self, value):
        """Camera acquisition time [s]."""
        params = self.parameters
        value = max(value, params.ACQUISITION_TIME_MIN)
        self['cam1:AcquirePeriod'] = value

    @property
    def acquisition_status(self):
        """Return IOC acquisition status."""
        return self['cam1:Acquire']

    @property
    def cam_max_sizex(self):
        """Camera max second dimension size [pixel]."""
        return self['cam1:MaxSizeX_RBV']

    @property
    def cam_max_sizey(self):
        """Camera max first dimension size [pixel]."""
        return self['cam1:MaxSizeY_RBV']

    @property
    def cam_sizex(self):
        """Camera image X width [pixels]."""
        # NOTE: the same as cam_sizex
        return self['cam1:SizeX_RBV']

    @cam_sizex.setter
    def cam_sizex(self, value):
        """Set camera image X width [pixel]."""
        # NOTE: acquisition has to be turned off and on for
        # this to take effect on ROI and image1 modules
        value = int(value)
        if 0 < self.cam_minx + value <= self.cam_max_sizex:
            self['cam1:SizeX'] = value
        else:
            raise ValueError('Invalid offsetx and width combination!')

    @property
    def cam_sizey(self):
        """Camera image Y height [pixels]."""
        # NOTE: the same as cam_sizey
        return self['cam1:SizeY_RBV']

    @cam_sizey.setter
    def cam_sizey(self, value):
        """Set camera image Y height [pixel]."""
        # NOTE: acquisition has to be turned off and on for
        # this to take effect on ROI and image1 modules
        value = int(value)
        if 0 < self.cam_miny + value <= self.cam_max_sizey:
            self['cam1:SizeY'] = value
        else:
            raise ValueError('Invalid offsety and height combination!')

    @property
    def cam_minx(self):
        """Camera image X offset [pixels]."""
        return self['cam1:MinX_RBV']

    @cam_minx.setter
    def cam_minx(self, value):
        """Set camera image X offset [pixel]."""
        value = int(value)
        if 0 < value + self.cam_sizex <= self.cam_max_sizex:
            self['cam1:MinX'] = value
        else:
            raise ValueError('Invalid offsetx and width combination!')

    @property
    def cam_miny(self):
        """Camera image Y offset [pixels]."""
        return self['cam1:MinY_RBV']

    @cam_miny.setter
    def cam_miny(self, value):
        """Set camera image Y offset [pixel]."""
        value = int(value)
        if 0 <= value + self.cam_sizey <= self.cam_max_sizey:
            self['cam1:MinY'] = value
        else:
            raise ValueError('Invalid offsety and height combination!')

    @property
    def cam_roi(self):
        """Return current ROI (offsetx, offsety, width, height)."""
        roi = (self.cam_minx, self.cam_miny, self.cam_sizex, self.cam_sizey)
        return roi

    @property
    def roi_minx(self):
        """."""
        return self['ROI1:MinX_RBV']

    @roi_minx.setter
    def roi_minx(self, value):
        """."""
        self['ROI1:MinX'] = int(value)

    @property
    def roi_miny(self):
        """."""
        return self['ROI1:MinY_RBV']

    @roi_miny.setter
    def roi_miny(self, value):
        """."""
        self['ROI1:MinY'] = int(value)

    @property
    def image_sizex(self):
        """Image second dimension size [pixel]."""
        return self['image1:ArraySize0_RBV']

    @property
    def image_sizey(self):
        """Image first dimension size [pixel]."""
        return self['image1:ArraySize1_RBV']

    @property
    def image(self):
        """Return DVF image formatted as a (sizey, sizex) numpy matrix."""
        shape = (self.image_sizey, self.image_sizex)
        data = self['image1:ArrayData']
        image = _np.reshape(data, shape)
        return image

    @property
    def image_auto_monitor(self):
        """Image PV auto monitor."""
        propty = self._get_propty('image1:ArrayData')
        return self.pv_object(propty).auto_monitor

    @image_auto_monitor.setter
    def image_auto_monitor(self, value):
        """Set image PV auto monitor."""
        propty = self._get_propty('image1:ArrayData')
        self.pv_object(propty).auto_monitor = bool(value)

    @property
    def image_pixel_size(self):
        """Image pixel size [um]."""
        params = self.parameters
        return params.IMAGE_PIXEL_SIZE

    @property
    def optics_magnification_factor(self):
        """Source to image magnification factor."""
        params = self.parameters
        return params.OPTICS_MAGNIFICATION_FACTOR

    @property
    def conv_pixel_2_srcsize(self):
        """Pixel to source size convertion factor."""
        pixel_size = self.image_pixel_size
        mag_factor = self.optics_magnefication_factor
        pixel2srcsize = pixel_size / mag_factor
        return pixel2srcsize

    @property
    def gain(self):
        """Return camera gain."""
        return self['cam1:Gain_RBV']

    @gain.setter
    def gain(self, value):
        """Set camera gain."""
        self['cam1:Gain'] = value

    @property
    def gain_auto(self):
        """Return camera gain auto."""
        return self['cam1:GainAuto_RBV']

    @gain_auto.setter
    def gain_auto(self, value):
        """Set camera gain auto."""
        self['cam1:GainAuto'] = value

    @property
    def pixel_format(self):
        """Return camera pixel format."""
        return self['cam1:PixelFormat_RBV']

    @pixel_format.setter
    def pixel_format(self, value):
        """Set camera pixel format."""
        self['cam1:PixelFormat'] = value

    @property
    def data_type(self):
        """Return image data type."""
        return self['image1:DataType_RBV']

    @property
    def cam_temperature(self):
        """Return camera temperature."""
        return self['cam1:GC_TemperatureAbs_RBV']

    @property
    def cam_frames_completed(self):
        """Return number of acquisition frames completed."""
        return self['cam1:ARFramesCompleted']

    @property
    def cam_frames_failures(self):
        """Return number of acquisition frames failures."""
        return self['cam1:ARFrameFailures']

    def cmd_reset(self, timeout=None):
        """Reset DVF to a standard configuration."""
        props_values = {
            'cam1:ArrayCallbacks': 1,  # Enable passing array
            'cam1:ImageMode': 2,  # Continuous
            'cam1:PixelFormat': 1,  # Mono12
            # ROI1 takes images from camera driver
            'ROI1:NDArrayPort': self['cam1:PortName_RBV'],
            'ROI1:EnableCallbacks': 1,  # Enable getting from NDArrayPort
            'ROI1:MinX': 0,  # [pixel]
            'ROI1:MinY': 0,  # [pixel]
            'ROI1:SizeX': self.cam_max_sizex,  # [pixel]
            'ROI1:SizeY': self.cam_max_sizey,  # [pixel]
            'ROI1:EnableX': 1,  # Enable
            'ROI1:EnableY': 1,  # Enable
            'ROI1:ArrayCallbacks': 1,  # Enable passing array
            # image1 takes images from ROI1
            'image1:NDArrayPort': self['ROI1:PortName_RBV'],
            'image1:EnableCallbacks': 1,  # Enable
            'Trans1:EnableCallbacks': 0,  # Disable
        }

        # set properties
        for propty, value in props_values.items():
            self[propty] = value

        # check readback values
        for propty, value in props_values.items():
            propty = self._get_propty(propty)
            if not self.wait(propty + '_RBV', value, timeout=timeout):
                return False

        # configure image acquisition parameters
        params = self.parameters
        self.exposure_time = params.EXPOSURE_TIME_DEFAULT
        self.acquisition_time = params.ACQUISITION_TIME_DEFAULT
        return True

    def cmd_acquire_on(self, timeout=_default_timeout):
        """Tune IOC image acquisition on."""
        return self._set_and_wait('cam1:Acquire', 1, timeout=timeout)

    def cmd_acquire_off(self, timeout=_default_timeout):
        """Tune IOC image acquisition off."""
        return self._set_and_wait('cam1:Acquire', 0, timeout=timeout)

    def cmd_cam_roi_set(self, offsetx, offsety, width, height, timeout=None):
        """Set cam image ROI and reset aquisition."""
        c_width, c_height = self.cam_sizex, self.cam_sizey
        n_width, n_height = int(width), int(height)

        if not self.cmd_acquire_off(timeout=timeout):
            return False
        if n_width < c_width:
            self._set_and_wait('cam1:SizeX', width, timeout=timeout)
            self._set_and_wait('cam1:MinX', offsetx, timeout=timeout)
        else:
            self._set_and_wait('cam1:MinX', offsetx, timeout=timeout)
            self._set_and_wait('cam1:SizeX', width, timeout=timeout)
        if n_height < c_height:
            self._set_and_wait('cam1:SizeY', height, timeout=timeout)
            self._set_and_wait('cam1:MinY', offsety, timeout=timeout)
        else:
            self._set_and_wait('cam1:MinY', offsety, timeout=timeout)
            self._set_and_wait('cam1:SizeY', height, timeout=timeout)
        if not self.cmd_acquire_on(timeout=timeout):
            return False

        return True

    def cmd_cam_roi_reset(self, timeout=None):
        """."""
        return self.cmd_cam_roi_set(
            offsetx=0,
            offsety=0,
            width=self.cam_max_sizex,
            height=self.cam_max_sizey,
            timeout=timeout,
        )

    def _get_propty(self, propty):
        if self.devname != DVF.DEVICES.BO_DVF:
            return propty

        if DVF.DEVICES.BO_DVF not in DVF._PROPTY_NAME_MAP:
            DVF._PROPTY_NAME_MAP[DVF.DEVICES.BO_DVF] = dict()
        propty_name_map = DVF._PROPTY_NAME_MAP[DVF.DEVICES.BO_DVF]
        if propty not in propty_name_map:
            if ':' in propty:
                plug, prop = propty.split(':')
                plug = plug.replace('cam', 'Cam')
                plug = plug.replace('image', 'Image')
                plug = plug.replace('Trans', 'Transf')
                plug = plug.replace('ARFrameFailures', 'FAILURES_RBV')
                plug = plug.replace('ARFramesCompleted', 'COMPLETED_RBV')
                plug = plug.replace('GC_TemperatureAbs_RBV', 'Temp-Mon')
                propty = plug + prop
            propty_name_map[propty] = propty
        return propty_name_map[propty]

    def __getitem__(self, propty):
        """Return value of property."""
        propty = self._get_propty(propty)
        return super().__getitem__(propty)

    def __setitem__(self, propty, value):
        """Set value of property."""
        propty = self._get_propty(propty)
        return super().__setitem__(propty, value)

    @staticmethod
    def conv_devname2parameters(devname):
        """."""
        return DVF._dev2params[devname]

    def _set_and_wait(self, propty, value, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        self[propty] = value
        propty = self._get_propty(propty)
        return self.wait(propty + '_RBV', value, timeout=timeout)


class DVFImgProc(DVF):
    """."""

    PROPERTIES_DEFAULT = DVF.PROPERTIES_DEFAULT + (
        'ImgIntensityMax-Mon',
        'ImgIntensityMin-Mon',
        'ImgIntensitySum-Mon',
        'ImgIsSaturated-Mon',
        'ImgIsWithBeam-Mon',
        'ImgIsWithBeamThreshold-SP',
        'ImgIsWithBeamThreshold-RB',
        'ImgROIX-RB',
        'ImgROIX-SP',
        'ImgROIXCenter-Mon',
        'ImgROIXFWHM-Mon',
        'ImgROIY-RB',
        'ImgROIY-SP',
        'ImgROIYCenter-Mon',
        'ImgROIYFWHM-Mon',
        'ImgLog-Mon',
        'ImgROIUpdateWithFWHM-Sel',
        'ImgROIUpdateWithFWHM-Sts',
        'ImgROIYUpdateWithFWHMFactor-RB',
        'ImgROIYUpdateWithFWHMFactor-SP',
        'ImgROIXUpdateWithFWHMFactor-RB',
        'ImgROIXUpdateWithFWHMFactor-SP',
        'ImgROIXFitMean-Mon',
        'ImgROIXFitSigma-Mon',
        'ImgROIXFitAmplitude-Mon',
        'ImgROIXFitError-Mon',
        'ImgROIYFitMean-Mon',
        'ImgROIYFitSigma-Mon',
        'ImgROIYFitAmplitude-Mon',
        'ImgROIYFitError-Mon',
        'ImgFitAngle-Mon',
        'ImgFitSigma1-Mon',
        'ImgFitSigma2-Mon',
        'ImgFitProcTime-Mon',
        'ImgFitAngleUseCMomSVD-Sel',
        'ImgFitAngleUseCMomSVD-Sts',
        'ImgDVFStatus-Mon',
        'ImgDVFStatusLabels-Cte',
    )

    def __init__(self, devname, props2init='all', **kwargs):
        """."""
        super().__init__(devname=devname, props2init=props2init, **kwargs)
        self._fitgaussian = _FitGaussianScipy()

    @property
    def intensity_min(self):
        """Image min intensity."""
        return self['ImgIntensityMin-Mon']

    @property
    def intensity_max(self):
        """Image max intensity."""
        return self['ImgIntensityMax-Mon']

    @property
    def intensity_sum(self):
        """Image sum intensity."""
        return self['ImgIntensitySum-Mon']

    @property
    def is_saturated(self):
        """Whether image is saturated."""
        return self['ImgIsSaturated-Mon']

    @property
    def is_with_beam(self):
        """Whether image is with beam."""
        return self['ImgIsWithBeam-Mon']

    @property
    def is_with_beam_threashold(self):
        """Get image is with beam threashold."""
        return self['ImgIsWithBeamThreshold-RB']

    @is_with_beam_threashold.setter
    def is_with_beam_threashold(self, value):
        """Set image is with beam threashold."""
        self['ImgIsWithBeamThreshold-SP'] = value

    @property
    def roiy(self):
        """."""
        return self['ImgROIY-RB']

    @roiy.setter
    def roiy(self, value):
        """."""
        self['ImgROIY-SP'] = value

    @property
    def roix(self):
        """."""
        return self['ImgROIX-RB']

    @roix.setter
    def roix(self, value):
        """."""
        self['ImgROIX-SP'] = value

    @property
    def roiy_center(self):
        """."""
        return self['ImgROIYCenter-Mon']

    @property
    def roix_center(self):
        """."""
        return self['ImgROIXCenter-Mon']

    @property
    def roiy_fwhm(self):
        """."""
        return self['ImgROIYFWHM-Mon']

    @property
    def roix_fwhm(self):
        """."""
        return self['ImgROIXFWHM-Mon']

    @property
    def roiy_fit_amplitude(self):
        """."""
        return self['ImgROIYFitAmplitude-Mon']

    @property
    def roix_fit_amplitude(self):
        """."""
        return self['ImgROIXFitAmplitude-Mon']

    @property
    def roiy_fit_mean(self):
        """."""
        return self['ImgROIYFitMean-Mon']

    @property
    def roix_fit_mean(self):
        """."""
        return self['ImgROIXFitMean-Mon']

    @property
    def roiy_fit_sigma(self):
        """."""
        return self['ImgROIYFitSigma-Mon']

    @property
    def roix_fit_sigma(self):
        """."""
        return self['ImgROIXFitSigma-Mon']

    @property
    def roiy_fit_error(self):
        """."""
        return self['ImgROIYFitError-Mon']

    @property
    def roix_fit_error(self):
        """."""
        return self['ImgROIXFitError-Mon']

    @property
    def roiy_fwhm_factor(self):
        """."""
        return self['ImgROIYUpdateWithFWHMFactor-RB']

    @roiy_fwhm_factor.setter
    def roiy_fwhm_factor(self, value):
        """."""
        self['ImgROIYUpdateWithFWHMFactor-SP'] = value

    @property
    def roix_fwhm_factor(self):
        """."""
        return self['ImgROIXUpdateWithFWHMFactor-RB']

    @roix_fwhm_factor.setter
    def roix_fwhm_factor(self, value):
        """."""
        self['ImgROIXUpdateWithFWHMFactor-SP'] = value

    @property
    def roi_update_with_fwhm(self):
        """."""
        return bool(self['ImgROIUpdateWithFWHM-Sts'])

    @roi_update_with_fwhm.setter
    def roi_update_with_fwhm(self, value):
        """."""
        self['ImgROIUpdateWithFWHM-Sel'] = bool(value)

    @property
    def fit_angle(self):
        """."""
        return self['ImgFitAngle-Mon']

    @property
    def fit_sigma1(self):
        """."""
        return self['ImgFitSigma1-Mon']

    @property
    def fit_sigma2(self):
        """."""
        return self['ImgFitSigma2-Mon']

    @property
    def fit_proctime(self):
        """Return image processing time [ms]."""
        return self['ImgFitProcTime-Mon']

    @property
    def fit_angle_use_cmom_svd(self):
        """."""
        return self['ImgFitAngleUseCMomSVD-Sts']

    @fit_angle_use_cmom_svd.setter
    def fit_angle_use_cmom_svd(self, value):
        """."""
        self['ImgFitAngleUseCMomSVD-Sel'] = bool(value)

    @property
    def log(self):
        """."""
        return self['ImgLog-Mon']

    def cam_roi_calc(self, roix_fwhm_factor, roiy_fwhm_factor):
        """Return ROI based on FWHM factors."""
        multp = 4

        width = 2 * self.roix_fwhm * abs(roix_fwhm_factor)
        width = int(width)
        width -= width % multp
        centerx = self.cam_minx + self.roix_center
        offsetx = int(centerx - width / 2)
        offsetx -= offsetx % multp

        height = 2 * self.roiy_fwhm * abs(roiy_fwhm_factor)
        height = int(height)
        height -= height % multp
        centery = self.cam_miny + self.roiy_center
        offsety = int(centery - height / 2)
        offsety -= offsety % multp

        params = (offsetx, offsety, width, height)

        # check parameters, return False if inconsistent
        if width <= 0 or width > self.cam_max_sizex:
            return False, params
        if height <= 0 or height > self.cam_max_sizey:
            return False, params
        if offsetx < 0 or offsetx + width >= self.cam_max_sizex:
            return False, params
        if offsety < 0 or offsety + height >= self.cam_max_sizey:
            return False, params

        # all ok, return parameters with True status.
        return True, params

    def create_image2dfit(self):
        """Return a Image2DFit object with current image as data."""
        imgfit2d = _Image2D_Fit(
            data=self.image,
            fitgauss=self._fitgaussian,
            roix=self.roix,
            roiy=self.roiy,
        )
        return imgfit2d
